/*
This program is inspired by the Tetris game ;)
Keyboard controls:

- - and = : to decrease/increase piece's angular velocity (creating fast rotating pieces can be fun ;)
- [ and ] : to decrease/increase piece's size
- ; and ' : to decrease/increase the piece merging frequency
- , and . : to decrease/increase the piece's average speed
- a and d : to rotate the "camera"

Tips: when increasing the merging frequency, try to also increase the average
speed, to get a more obvious effect.
*/

/* Sizes */
int WindowWidth = 800;
int WindowHeight = 800;
float PieceSize = 20;
float OuterWindowWidth = WindowWidth + 2 * PieceSize;
float OuterWindowHeight = WindowHeight + 2 * PieceSize;

/* Piece position */
int NumPieces = 64; // try increasing this to a big number, say, 512, for a different feel ;)
                    // it's possible to provide control for this, but i'm just too lazy now
float[] PieceStartX = new float[NumPieces];
float[] PieceStartY = new float[NumPieces];
float[] PieceEndX = new float[NumPieces];
float[] PieceEndY = new float[NumPieces];
float[] PieceX = new float[NumPieces];
float[] PieceY = new float[NumPieces];
float[] PieceBlockX = new float[NumPieces]; // the x destination of piece i when it belongs to some block
float[] PieceBlockY = new float[NumPieces];

/* Piece motion */
float[] PieceSpeed = new float[NumPieces];
float PieceSpeedAverage = 0.004;
float[] A = new float[NumPieces];
float[] K = new float[NumPieces];
float[] T = new float[NumPieces];

/* Piece orientation */
float[] PieceAngle = new float[NumPieces];
float[] PieceAngleStep = new float[NumPieces];
float PieceAngularVelocity = 80;

/* Piece-block */
// PieceToBlockIndex[i] = the block piece i belongs to
int[] PieceToBlockIndex = new int[NumPieces]; 

/* Piece merging */
float MergeFrequency = 0.02;

/* Piece states */
int PIECE_FREE = 0;
int PIECE_CHOSEN_TO_FORM_BLOCK = 1;
int PIECE_MOVING_TO_BLOCK = 2;
int PIECE_IN_BLOCK_WAITING = 3;
int PIECE_IN_BLOCK_MOVING = 4;
int[] PieceState = new int[NumPieces];
int NumFreePieces = NumPieces;

/* Piece color */
color[] PieceColor = new color[NumPieces];
color[] PieceColorFinal = new color[NumPieces];

/* Number of blocks */
int NumBlocksMax = NumPieces / 4;
int NumBlocks = 0;

/* Block types */
int BLOCK_O_SHAPE = 0;
int BLOCK_I_SHAPE = 1;
int BLOCK_L_SHAPE = 2;
int BLOCK_T_SHAPE = 3;
int BLOCK_Z_SHAPE = 4;
// TODO: J shape and S shape
int NUM_BLOCK_TYPES_MAX = 5;
int[] BlockType = new int[NumBlocksMax];

/* Block position */
float[] BlockX = new float[NumBlocksMax];
float[] BlockY = new float[NumBlocksMax];
float[] BlockStartX = new float[NumBlocksMax];
float[] BlockStartY = new float[NumBlocksMax];
float[] BlockEndX = new float[NumBlocksMax];
float[] BlockEndY = new float[NumBlocksMax];

/* Block-piece */
int[] NumPiecesPerBlock = new int[NumBlocksMax];
int[] BlockToPieceIndex0 = new int[NumBlocksMax]; // BlockToPieceIndex0[i] = the first piece of block i
int[] BlockToPieceIndex1 = new int[NumBlocksMax]; 
int[] BlockToPieceIndex2 = new int[NumBlocksMax];
int[] BlockToPieceIndex3 = new int[NumBlocksMax];

/* Block motion */
float[] BlockT = new float[NumBlocksMax];
float[] BlockSpeed = new float[NumBlocksMax];
int[] BlockFadingFrame = new int[NumBlocksMax];
int BLOCK_FADING_NUM_FRAMES_MAX = 10;

/* Block states */
int BLOCK_NOT_EXIST = 0;
int BLOCK_WAITING_FOR_PIECES = 1;
int BLOCK_FADING_IN = 2;
int BLOCK_MOVING = 3;
int BLOCK_FADING_OUT = 4;
int[] BlockState = new int[NumBlocksMax];

/* Block color */
color[] BlockColor = new color[NumBlocksMax];
color[] BlockColorFinal = new color[NumBlocksMax];

/* Block size and orientation */
float[] BlockSizeMultiplier = new float[NumBlocksMax];
float BLOCK_SIZE_STEP = 0.01;
float[] BlockAngle = new float[NumBlocksMax];
float[] BlockAngleDeltaLastFrame = new float[NumBlocksMax];

/* Camera */
float CamRotate = 0;
float CamRotateSpeed = 1.0;

/* Some colors */
int AlphaFree  = 80;
int AlphaBlock = 200; 
color WHITE = color(255, 255, 255, 255);

void setup() {  
  size(800, 800);
  smooth();
  initialize();
}

void initialize() {
  for (int i = 0; i < NumBlocksMax; ++i) {  
    BlockSizeMultiplier[i] = 1;
  }
  
  restartAllPieces();
}

void restartAllPieces() {
  for (int i = 0; i < NumPieces; ++i) {
    restartPiece(i);
  }
}

void restartPiece(int i) {
  setPiecePositions(i);    
  setPieceColor(i);
  setPieceMotion(i);
  setPieceOrientation(i);
}

void setPiecePositions(int i) {
  setPieceStartPosition(i);
  setPieceEndPosition(i);    
  /* set the piece's current position, relative to StartX, StartY */
  PieceX[i] = 0;
  PieceY[i] = 0;
}

void setPieceStartPosition(int i) {
  float[] randomStart = GetRandomPointOnOuterWindowEdge();
  PieceStartX[i] = randomStart[0];
  PieceStartY[i] = randomStart[1];
}

void setPieceEndPosition(int i) {
  if (PieceState[i] == PIECE_FREE) {
    float[] randomEnd = GetRandomPointOnOuterWindowEdge();
    PieceEndX[i] = randomEnd[0];
    PieceEndY[i] = randomEnd[1];
  } else if (PieceState[i] == PIECE_CHOSEN_TO_FORM_BLOCK) {
    PieceState[i] = PIECE_MOVING_TO_BLOCK;
    PieceEndX[i] = PieceBlockX[i];
    PieceEndY[i] = PieceBlockY[i];
  }
}

void setPieceColor(int i) {
  colorMode(HSB);
  PieceColor[i] = PieceColorFinal[i] = color(random(0, 127), random(180, 255), random(180, 255), AlphaFree);
  colorMode(RGB);
  //PieceColor[i] = PieceColorFinal[i] = color(0, 255, 0, Alpha);
}

void setPieceMotion(int i) {  
  PieceSpeed[i] = random(PieceSpeedAverage - 0.002, PieceSpeedAverage + 0.002);
  A[i] = random(1, 300);
  K[i] = random(1, 1.1);
  T[i] = 0;
}

/* Here we basically aim to compute an "angle step" (in radians) that is added to the
   piece's orientation each frame, so that when the piece arrives at its destination, it is
   exactly upright. I'm not sure if the formula here works correctly, but the animation
   looks good enough to me ;) */
void setPieceOrientation(int i) {
  float numSteps = 1 / PieceSpeed[i];
  float dstAngle = GetAngleWithXAxis(PieceStartX[i], PieceStartY[i], PieceEndX[i], PieceEndY[i]);  
  dstAngle = degrees(dstAngle + PI);
  float srcAngle = random(0, 360);
  PieceAngle[i] = radians(srcAngle);
  float numFrames = 1 / PieceSpeed[i];
  float numRounds = dist(PieceStartX[i], PieceStartY[i], PieceEndX[i], PieceEndY[i]) /
                    random(PieceAngularVelocity - 20, PieceAngularVelocity + 20);
  float step = random(0.0, 1.0) <= 0.5 ? (dstAngle + numRounds * 360 - srcAngle) / numFrames
                                       : (dstAngle - numRounds * 360 - srcAngle) / numFrames;  
  PieceAngleStep[i] = radians(step);  
}

float[] GetRandomPointOnOuterWindowEdge() {
  float x, y;
  float outerWindowCircumference = 2 * (OuterWindowWidth + OuterWindowHeight);
  float pos = random(0, outerWindowCircumference);
  if (pos < OuterWindowWidth) { // top edge
    x = pos;
    y = 0;
  } else if (pos < OuterWindowWidth + OuterWindowHeight) { // right edge
    x = OuterWindowWidth;
    y = pos - OuterWindowWidth;
  } else if (pos < 2 * OuterWindowWidth + OuterWindowHeight) { // bottom edge
    x = OuterWindowWidth - (pos - OuterWindowWidth - OuterWindowHeight);
    y = OuterWindowHeight;
  } else { // left edge
    x = 0;
    y = OuterWindowHeight - (pos - 2 * OuterWindowWidth - OuterWindowHeight);
  }  
  return new float[]{x - PieceSize, y - PieceSize};
}

void draw() {
  handleCamera();
  handleAllPieces();  
  mergePieces();
  handleAllBlocks();
  background(40);
  drawAllBlocks();
  drawAllPieces();
}

void handleCamera() {
  translate(WindowWidth/2, WindowHeight/2);
  rotate(radians(CamRotate));  
  translate(-WindowWidth/2, -WindowHeight/2);
}

void handleAllPieces() {
  for (int i = 0; i < NumPieces; ++i) {
    handlePiece(i);    
  }
}

void handlePiece(int i) {
  switch (PieceState[i]) {
    case 0: // PIECE_FREE (we can't switch on non-constant variables such as PIECE_FREE...)
      handlePieceFree(i);
      break;
    case 1: // PIECE_CHOSEN_TO_FORM_BLOCK
      handlePieceChosenToFormBlock(i);
      break;
    case 2: // PIECE_MOVING_TO_BLOCK
      handlePieceMovingToBlock(i);
      break;
    case 3: // PIECE_IN_BLOCK_WAITING
      handlePieceInBlockWaiting(i);
      break;
    case 4: // PIECE_IN_BLOCK_MOVING
      handlePieceInBlockMoving(i);
      break;
  }
}

void handlePieceFree(int i) {
  movePiece(i);
  
  if (T[i] > 1) {
    restartPiece(i);
  }    
}

void movePiece(int i) {
  PieceX[i] = map(T[i], 0.0, 1.0, 0.0, dist(PieceStartX[i], PieceStartY[i], PieceEndX[i], PieceEndY[i]));
  PieceY[i] = A[i] * sin(2 * PI * T[i]);
  T[i] += PieceSpeed[i];
  PieceAngle[i] += PieceAngleStep[i];
}

void handlePieceChosenToFormBlock(int i) {
  handlePieceFree(i);
}

void handlePieceMovingToBlock(int i) {
  movePiece(i);
  
  if (T[i] > 1) {
    ++NumPiecesPerBlock[PieceToBlockIndex[i]];
    PieceState[i] = PIECE_IN_BLOCK_WAITING;
    PieceX[i] = PieceBlockX[i];
    PieceY[i] = PieceBlockY[i];
    //noLoop();
  }   
  
  //PieceColor[i] = color(255, 0, 0, Alpha);
}

void handlePieceInBlockWaiting(int i) {  
  int t = BlockFadingFrame[PieceToBlockIndex[i]];
  int h = BLOCK_FADING_NUM_FRAMES_MAX / 2;  
  if (t <  h) {
    PieceColor[i] = lerpColor(PieceColorFinal[i], WHITE, float(t) / h);
  }
  
  //PieceColor[i] = color(0, 0, 255, Alpha);
}

void handlePieceInBlockMoving(int i) {
  // do nothing
}

void mergePieces() {
  if ((NumBlocks >= NumBlocksMax) || (random(0.0, 1.0) > MergeFrequency)) {
    return;
  }  
  
  int blockIndex = getNewBlockIndex();
  if (blockIndex == -1) { // too many blocks existing
    return;
  }
    
  int[] pieces = Get4RandomFreePieces(4, NumPieces);
  if (pieces[0] == -1) { // no free pieces left
    return;
  }
  
  NumFreePieces -= 4;
  
  BlockType[blockIndex] = int(random(0, NUM_BLOCK_TYPES_MAX));
  //BlockType[blockIndex] = BLOCK_Z_SHAPE;
  
  setPieceBlockState(pieces);
  float[] blockPos = GetRandomPositionInWindow();
  setPieceBlockPosition(BlockType[blockIndex], blockPos, pieces);  
  setPieceToBlockIndex(blockIndex, pieces);
  
  setBlockPosition(blockIndex, blockPos);
  NumPiecesPerBlock[blockIndex] = 0;
  setBlockToPieceIndex(blockIndex, pieces);
  BlockState[blockIndex] = BLOCK_WAITING_FOR_PIECES;
  NumBlocks++;
  
  setBlockColor(blockIndex); 

  BlockAngleDeltaLastFrame[blockIndex] = random(-1, 1);
}

void setPieceBlockPosition(int blockType, float[] blockPos, int[] pieces) {  
  float[] piecePos0 = GetBlockPiecesDestinations(blockType, 0, blockPos);
  float[] piecePos1 = GetBlockPiecesDestinations(blockType, 1, blockPos);
  float[] piecePos2 = GetBlockPiecesDestinations(blockType, 2, blockPos);
  float[] piecePos3 = GetBlockPiecesDestinations(blockType, 3, blockPos);
  PieceBlockX[pieces[0]] = piecePos0[0]; PieceBlockY[pieces[0]] = piecePos0[1];
  PieceBlockX[pieces[1]] = piecePos1[0]; PieceBlockY[pieces[1]] = piecePos1[1];
  PieceBlockX[pieces[2]] = piecePos2[0]; PieceBlockY[pieces[2]] = piecePos2[1];
  PieceBlockX[pieces[3]] = piecePos3[0]; PieceBlockY[pieces[3]] = piecePos3[1];     
}

void setPieceBlockState(int[] pieces) {
  PieceState[pieces[0]] = PIECE_CHOSEN_TO_FORM_BLOCK;
  PieceState[pieces[1]] = PIECE_CHOSEN_TO_FORM_BLOCK;
  PieceState[pieces[2]] = PIECE_CHOSEN_TO_FORM_BLOCK;
  PieceState[pieces[3]] = PIECE_CHOSEN_TO_FORM_BLOCK;
}

void setPieceToBlockIndex(int blockIndex, int[] pieces) {
  PieceToBlockIndex[pieces[0]] = blockIndex;
  PieceToBlockIndex[pieces[1]] = blockIndex;
  PieceToBlockIndex[pieces[2]] = blockIndex;
  PieceToBlockIndex[pieces[3]] = blockIndex;
}

void setBlockPosition(int blockIndex, float[] blockPos) {
  BlockX[blockIndex] = blockPos[0];
  BlockY[blockIndex] = blockPos[1];
}

void setBlockToPieceIndex(int blockIndex, int[] pieces) {
  BlockToPieceIndex0[blockIndex] = pieces[0];
  BlockToPieceIndex1[blockIndex] = pieces[1];
  BlockToPieceIndex2[blockIndex] = pieces[2];
  BlockToPieceIndex3[blockIndex] = pieces[3];
}

void setBlockColor(int i) {
  colorMode(HSB);
  BlockColor[i] = BlockColorFinal[i] = color(random(127, 255), random(180, 255), random(180, 255));
  colorMode(RGB);
}

/* pieceNumber is sorted top to bottom, left to right */
float[] GetBlockPiecesDestinations(int blockType, int pieceNumber, float[] blockPos) {
  switch (blockType) {
    case 0: // BLOCK_O_SHAPE
      return GetPiecesDestinationsOShape(pieceNumber, blockPos);      
    case 1: // BLOCK_I_SHAPE
      return GetPiecesDestinationsIShape(pieceNumber, blockPos);      
    case 2: // BLOCK_L_SHAPE
      return GetPiecesDestinationsLShape(pieceNumber, blockPos);      
    case 3: // BLOCK_T_SHAPE
      return GetPiecesDestinationsTShape(pieceNumber, blockPos);      
    case 4: // BLOCK_Z_SHAPE
      return GetPiecesDestinationsZShape(pieceNumber, blockPos);      
  }
  return new float[] { -1 };
}

float[] GetPiecesDestinationsOShape(int pieceNumber, float[] blockPos) {
  float[] results = new float[2];
  switch(pieceNumber) {
  case 0:
    results[0] = blockPos[0] - PieceSize / 2;
    results[1] = blockPos[1] - PieceSize / 2;
    break;
  case 1:
    results[0] = blockPos[0] + PieceSize / 2;
    results[1] = blockPos[1] - PieceSize / 2;
    break;
  case 2:
    results[0] = blockPos[0] - PieceSize / 2;
    results[1] = blockPos[1] + PieceSize / 2;
    break;
  case 3:
    results[0] = blockPos[0] + PieceSize / 2;
    results[1] = blockPos[1] + PieceSize / 2;
    break;
  }
  return results;
}

float[] GetPiecesDestinationsIShape(int pieceNumber, float[] blockPos) {
  float[] results = new float[2];
  switch(pieceNumber) {
  case 0:
    results[0] = blockPos[0];
    results[1] = blockPos[1] - 3 * PieceSize / 2;
    break;
  case 1:
    results[0] = blockPos[0];
    results[1] = blockPos[1] - PieceSize / 2;
    break;
  case 2:
    results[0] = blockPos[0];
    results[1] = blockPos[1] + PieceSize / 2;
    break;
  case 3:
    results[0] = blockPos[0];
    results[1] = blockPos[1] + 3 * PieceSize / 2;
    break;
  }
  return results;
}

float[] GetPiecesDestinationsLShape(int pieceNumber, float[] blockPos) {
  float[] results = new float[2];
  switch(pieceNumber) {
  case 0:
    results[0] = blockPos[0] - PieceSize / 2;
    results[1] = blockPos[1] - PieceSize;
    break;
  case 1:
    results[0] = blockPos[0] - PieceSize / 2;
    results[1] = blockPos[1];
    break;
  case 2:
    results[0] = blockPos[0] - PieceSize / 2;
    results[1] = blockPos[1] + PieceSize;
    break;
  case 3:
    results[0] = blockPos[0] + PieceSize / 2;
    results[1] = blockPos[1] + PieceSize;
    break;
  }
  return results;
}

float[] GetPiecesDestinationsTShape(int pieceNumber, float[] blockPos) {
  float[] results = new float[2];
  switch(pieceNumber) {
  case 0:
    results[0] = blockPos[0];
    results[1] = blockPos[1] - PieceSize / 2;
    break;
  case 1:
    results[0] = blockPos[0] - PieceSize;
    results[1] = blockPos[1] + PieceSize / 2;
    break;
  case 2:
    results[0] = blockPos[0];
    results[1] = blockPos[1] + PieceSize / 2;
    break;
  case 3:
    results[0] = blockPos[0] + PieceSize;
    results[1] = blockPos[1] + PieceSize / 2;
    break;
  }
  return results;
}

float[] GetPiecesDestinationsZShape(int pieceNumber, float[] blockPos) {
  float[] results = new float[2];
  switch(pieceNumber) {
  case 0:
    results[0] = blockPos[0] - PieceSize;
    results[1] = blockPos[1] - PieceSize / 2;
    break;
  case 1:
    results[0] = blockPos[0];
    results[1] = blockPos[1] - PieceSize / 2;
    break;
  case 2:
    results[0] = blockPos[0];
    results[1] = blockPos[1] + PieceSize / 2;
    break;
  case 3:
    results[0] = blockPos[0] + PieceSize;
    results[1] = blockPos[1] + PieceSize / 2;
    break;
  }
  return results;
}

int getNewBlockIndex() {
  for (int i = 0; i < NumBlocksMax; ++i) {
    if (BlockState[i] == BLOCK_NOT_EXIST) {
      return i;
    }
  }
  return -1;
}

/* This is a modified version of Knuth's algorithm to find m different random numbers from
0 to n-1. In our case, not only the 4 pieces must be different, they must also be free. */
int[] Get4RandomFreePieces(int m, int n) {
  int[] results = new int[m];
  int j = -1;
  int k = 0;
  for (int i = 0; i < n; ++i) {
    if (PieceState[i] != PIECE_FREE) {
      continue;
    } else {
      ++j;
    }
    
    if (int(random(0, NumFreePieces - j)) < m) { // choose i with probability (NumFreePieces-j)/m
      results[k++] = i;
      --m;
    }       
  }
  return results;
}

float[] GetRandomPositionInWindow() {
  return new float[]{ random(0, WindowWidth), random(0, WindowHeight) };
}

void handleAllBlocks() {
  for (int i = 0; i < NumBlocksMax; ++i) {
    handleBlock(i);
  }
}

void handleBlock(int i) {
  /* The multiple if statements have the effect of a "smooth fall-through",
     that is, we don't have to do extra book keeping at the end of a state before
     transitioning to a new state, just to keep things consistent and avoid disynchronization
     with draw(). If these ifs are replaced with a switch, there will be flickerings now and then. */
  
  if (BlockState[i] == BLOCK_NOT_EXIST) {
    handleBlockNotExist(i);
  }
  
  if (BlockState[i] == BLOCK_WAITING_FOR_PIECES) {
    handleBlockWaitingForPieces(i);
  }
  
  if (BlockState[i] == BLOCK_FADING_IN) {
    handleBlockFadingIn(i);
  }
  
  if (BlockState[i] == BLOCK_MOVING) {
    handleBlockMoving(i);
  }
  
  if (BlockState[i] == BLOCK_FADING_OUT) {
    handleBlockFadingOut(i);
  }    
}

void handleBlockNotExist(int i) {
  // do nothing of course
}

void handleBlockWaitingForPieces(int i) {
  if (NumPiecesPerBlock[i] >= 4) {
    BlockState[i] = BLOCK_FADING_IN;
  }
}

void handleBlockFadingIn(int i) {
  if (BlockFadingFrame[i] >= BLOCK_FADING_NUM_FRAMES_MAX) {
      BlockFadingFrame[i] = 0;
      BlockState[i] = BLOCK_MOVING;
      PieceState[BlockToPieceIndex0[i]] = PIECE_IN_BLOCK_MOVING;
      PieceState[BlockToPieceIndex1[i]] = PIECE_IN_BLOCK_MOVING;
      PieceState[BlockToPieceIndex2[i]] = PIECE_IN_BLOCK_MOVING;
      PieceState[BlockToPieceIndex3[i]] = PIECE_IN_BLOCK_MOVING;      
      setBlockStartEnd(i);
  } else {
    int h = BLOCK_FADING_NUM_FRAMES_MAX / 2;
    if (BlockFadingFrame[i] >= h) {
      BlockColor[i] = lerpColor(WHITE, BlockColorFinal[i], float(BlockFadingFrame[i] - h) / h);
    }
    ++BlockFadingFrame[i];
  }
}

void setBlockStartEnd(int i) {
  float[] dest = GetRandomPositionInWindow();
  BlockEndX[i] = dest[0];
  BlockEndY[i] = dest[1];
  BlockSpeed[i] = random(0.005, 0.01);
  BlockStartX[i] = BlockX[i];
  BlockStartY[i] = BlockY[i];
}

void handleBlockMoving(int i) {
  if (BlockT[i] < 1) {
    BlockX[i] = map(BlockT[i], 0.0, 1.0, 0.0,
                    dist(BlockStartX[i], BlockStartY[i], BlockEndX[i], BlockEndY[i]));
    BlockY[i] = 0;    
    BlockColor[i] = color(red(BlockColorFinal[i]), green(BlockColorFinal[i]), blue(BlockColorFinal[i]), lerp(AlphaBlock, 0, BlockT[i]));
    BlockSizeMultiplier[i] += BLOCK_SIZE_STEP;
    float delta = sign(BlockAngleDeltaLastFrame[i]) * random(PI / 30);    
    BlockAngle[i] += delta;
    BlockAngleDeltaLastFrame[i] = delta;
    BlockT[i] += BlockSpeed[i];
  } else {
    BlockState[i] = BLOCK_FADING_OUT;
    BlockT[i] = 0;
  }
}

int sign(float x) {
  if (x >= 0) { // not totally correct, but we prefer it here :)
    return 1;
  } else {
    return -1;
  }
}

void handleBlockFadingOut(int i) {  
  discardBlock(i);    
}

void discardBlock(int i) {
  BlockState[i] = BLOCK_NOT_EXIST;
  NumPiecesPerBlock[i] = 0;
  --NumBlocks;
  BlockSizeMultiplier[i] = 1;
  BlockAngle[i] = 0;
  BlockAngleDeltaLastFrame[i] = random(-1, 1);
  releasePiecesFromBlock(i);  
}

void releasePiecesFromBlock(int i) {
  PieceState[BlockToPieceIndex0[i]] = PIECE_FREE;
  PieceState[BlockToPieceIndex1[i]] = PIECE_FREE;
  PieceState[BlockToPieceIndex2[i]] = PIECE_FREE;
  PieceState[BlockToPieceIndex3[i]] = PIECE_FREE;
  restartPiece(BlockToPieceIndex0[i]);
  restartPiece(BlockToPieceIndex1[i]);
  restartPiece(BlockToPieceIndex2[i]);
  restartPiece(BlockToPieceIndex3[i]);
}

void drawAllBlocks() {
  for (int i = 0; i < NumBlocksMax; ++i) {
    drawBlock(i);
  }
}

void drawBlock(int i) {
  switch (BlockState[i]) {
    case 0: // BLOCK_NOT_EXIST
      drawBlockNotExist(i);
      break;
    case 1: // BLOCK_WAITING_FOR_PIECES
      drawBlockWaitingForPieces(i);
      break;
    case 2: // BLOCK_FADING_IN
      drawBlockFadingIn(i);
      break;
    case 3: // BLOCK_MOVING
      drawBlockMoving(i);
      break;
    case 4: // BLOCK_FADING_OUT
      drawBlockFadingOut(i);
      break;
  }
}

void drawBlockNotExist(int i) {
  // do nothing
}

void drawBlockWaitingForPieces(int i) {
  // do nothing  
  noFill();
  stroke(255, 255, 255, 50);
  pushMatrix();
    translate(BlockX[i], BlockY[i]);
    drawBasicBlock(i);
  popMatrix();
  
}

void drawBlockFadingIn(int i) {
  stroke(255, 255, 255, alpha(BlockColor[i]));
  fill(BlockColor[i]);
  pushMatrix();
    translate(BlockX[i], BlockY[i]);
    drawBasicBlock(i);
  popMatrix();
}

void drawBlockMoving(int i) {
  stroke(255, 255, 255, alpha(BlockColor[i]));
  fill(BlockColor[i]);
  pushMatrix();
    translate(BlockStartX[i], BlockStartY[i]);
    float angle = GetAngleWithXAxis(BlockStartX[i], BlockStartY[i], BlockEndX[i], BlockEndY[i]);
    rotate(angle);
    translate(BlockX[i], BlockY[i]);
    rotate(-angle);
    scale(BlockSizeMultiplier[i]);
    rotate(BlockAngle[i]);
    drawBasicBlock(i);
  popMatrix();
}

void drawBlockFadingOut(int i) {
  // The block actually fades out while it's moving, so this function is
  // somewhat redundant. But it could be useful if we later decide to implement
  // some visual effect at the end of each block's life.
  drawBlockMoving(i);
}

void drawBasicBlock(int i) {
  switch (BlockType[i]) {
    case 0: // BLOCK_O_SHAPE
      drawBasicBlockOShape();
      break;
    case 1: // BLOCK_I_SHAPE
      drawBasicBlockIShape();
      break;
    case 2: // BLOCK_L_SHAPE
      drawBasicBlockLShape();
      break;
    case 3: // BLOCK_T_SHAPE
      drawBasicBlockTShape();
      break;
    case 4: // BLOCL_Z_SHAPE
      drawBasicBlockZShape();
      break;
  }
}

void drawBasicBlockOShape() {
  rectMode(CENTER);
  rect(0, 0, PieceSize * 2, PieceSize * 2);
  rectMode(CORNER);
}

void drawBasicBlockIShape() {
  rectMode(CENTER);
  rect(0, 0, PieceSize, PieceSize * 4);
  rectMode(CORNER);
}

void drawBasicBlockLShape() {
  beginShape();
    vertex(-PieceSize, -3 * PieceSize / 2);
    vertex(0, -3 * PieceSize / 2);
    vertex(0, PieceSize / 2);
    vertex(PieceSize, PieceSize / 2);
    vertex(PieceSize, 3 * PieceSize / 2);
    vertex(-PieceSize, 3 * PieceSize / 2);
  endShape(CLOSE);
}

void drawBasicBlockTShape() {
  beginShape();
    vertex(-PieceSize / 2, -PieceSize);
    vertex(PieceSize / 2, -PieceSize);
    vertex(PieceSize / 2, 0);
    vertex(3 * PieceSize / 2, 0);
    vertex(3 * PieceSize / 2, PieceSize);
    vertex(-3 * PieceSize / 2, PieceSize);
    vertex(-3 * PieceSize / 2, 0);
    vertex(-PieceSize / 2, 0);
  endShape(CLOSE);
}

void drawBasicBlockZShape() {
  beginShape();
    vertex(-3 * PieceSize / 2, -PieceSize);
    vertex(PieceSize / 2, -PieceSize);
    vertex(PieceSize / 2, 0);
    vertex(3 * PieceSize / 2, 0);
    vertex(3 * PieceSize / 2, PieceSize);
    vertex(-PieceSize / 2, PieceSize);
    vertex(-PieceSize / 2, 0);
    vertex(-3 * PieceSize / 2, 0);
  endShape(CLOSE);
}

void drawAllPieces() {
  for (int i = 0; i < NumPieces; ++i) {
    drawPiece(i);
  }  
}

void drawPiece(int i) {  
  switch (PieceState[i]) {
    case 0: // PIECE_FREE
      drawPieceFree(i);
      break;
    case 1: // PIECE_CHOSEN_TO_FORM_BLOCK
      drawPieceChosenToFormBlock(i);
      break;
    case 2: // PIECE_MOVING_TO_BLOCK
      drawPieceMovingToBlock(i);
      break;
    case 3: // PIECE_IN_BLOCK_WAITING
      drawPieceInBlockWaiting(i);
      break;
    case 4: // PIECE_IN_BLOCK_MOVING
      drawPieceInBlockMoving(i);
      break;
  }
}

void drawPieceFree(int i) {
  stroke(255, 255, 255, AlphaFree);
  fill(PieceColor[i]);
  pushMatrix();
    translate(PieceStartX[i], PieceStartY[i]);
    float angle = GetAngleWithXAxis(PieceStartX[i], PieceStartY[i], PieceEndX[i], PieceEndY[i]);
    rotate(angle);
    translate(PieceX[i], PieceY[i]);    
    rotate(PieceAngle[i]);
    drawBasicPiece(i);
  popMatrix();
}

void drawPieceChosenToFormBlock(int i) {
  drawPieceFree(i);
}

void drawPieceMovingToBlock(int i) {
  drawPieceFree(i);
}

void drawPieceInBlockWaiting(int i) {
  stroke(255, 255, 255, AlphaBlock);
  fill(PieceColor[i], AlphaBlock);
  pushMatrix();
    translate(PieceX[i], PieceY[i]);
    drawBasicPiece(i);
  popMatrix();
}

void drawPieceInBlockMoving(int i) {
  // do nothing
}

void drawBasicPiece(int i) {
  rectMode(CENTER);
  rect(0, 0, PieceSize, PieceSize);
  rectMode(CORNER);
}

/* Returns the angle that rotates the x axis clockwise to the vector going
from (x1, y1) to (x2, y2). The returned value is between -PI and PI. */
float GetAngleWithXAxis(float x1, float y1, float x2, float y2) {
  float dx = x2 - x1;
  float dy = y2 - y1;
  float dotProduct = dx * 1 + dy * 0; // (dx, dy) . (1, 0)
  float vecLength = sqrt(sq(dx) + sq(dy));  
  float phi = acos(dotProduct / vecLength);
  if (dy < 0) {
    return -phi;
  } else {
    return phi;
  }
}

void keyPressed() {
  if (key == 'a') {
    CamRotate += CamRotateSpeed;
    if (CamRotate > 360) {
      CamRotate = 0;
    }
  } else if (key == 'd') {
    CamRotate -= CamRotateSpeed;
    if (CamRotate < 0) {
      CamRotate = 360;
    }
  } else if (key == ']') {
    PieceSize += 10;
  } else if (key == '[') {
    PieceSize -= 10;
    PieceSize = PieceSize > 0 ? PieceSize : PieceSize + 10;
  } else if (key == '\'') {
    MergeFrequency += 0.001;
    MergeFrequency = MergeFrequency <= 1 ? MergeFrequency : MergeFrequency - 0.001;
    println("Merge frequency = " + MergeFrequency);
  } else if (key == ';') {
    MergeFrequency -= 0.001;
    MergeFrequency = MergeFrequency >= 0 ? MergeFrequency : MergeFrequency + 0.001;
    println("Merge frequency = " + MergeFrequency);
  } else if (key == '.') {
    PieceSpeedAverage += 0.001;
    println("Average speed = " + PieceSpeedAverage);
  } else if (key == ',') {
    PieceSpeedAverage -= 0.001;
    PieceSpeedAverage = PieceSpeedAverage > 0 ? PieceSpeedAverage : PieceSpeedAverage + 0.001;
    println("Average speed = " + PieceSpeedAverage);
  } else if (key == '-') {
    PieceAngularVelocity += 10;;
    println("Average angular velocity = " + PieceAngularVelocity);
  } else if (key == '=') {
    PieceAngularVelocity -= 10;;
    PieceAngularVelocity = PieceAngularVelocity >= 0 ? PieceAngularVelocity : PieceAngularVelocity + 10;
    println("Average angular velocity = " + PieceAngularVelocity);    
  }
}