import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class Week6_Tool extends PApplet {

/*
This program helps you design objects using Bezier curves.
The controls are:
- Left click on a point to select it
- Hold shift to select another point
- Press p when selecting two adjacent points will create a new point between them
- Press r/R to change the size of the control points.
*/

int WindowWidth = 800;
int WindowHeight = 800;
int CenterX = WindowWidth / 2;
int CenterY = WindowHeight / 2;

int NumPoints = 4;
float[] PointsX = new float[NumPoints];
float[] PointsY = new float[NumPoints];

float PointDiameter = 20;

int SelectedPoint1 = -1;
int SelectedPoint2 = -1;
int DraggedPoint = -1;
boolean Dragging = false;
float DraggedPointStartX = 0;
float DraggedPointStartY = 0;

public void setup() {
  smooth();
  size(WindowWidth, WindowHeight);
  InitRandomCurve();  
}

public void InitRandomCurve() {
  for (int i = 0; i < PointsX.length; ++i) {
    PointsX[i] = random(0, WindowWidth);
    PointsY[i] = random(0, WindowHeight);
  }
}

public void mousePressed() {
  int closestPoint = GetClosestPointToMouse();     
  if (keyPressed && key == CODED && keyCode == SHIFT) {
    if (SelectedPoint1 == -1) {
      SelectedPoint1 = closestPoint;
    } else {
      SelectedPoint2 = closestPoint;
    }
  } else {
    SelectedPoint1 = closestPoint;
    SelectedPoint2 = -1;
  }
  
  if (closestPoint != -1) {
    DraggedPoint = closestPoint;
    Dragging = true;
    DraggedPointStartX = mouseX;
    DraggedPointStartY = mouseY;
    println("Point " + closestPoint);
    print("x = " + PointsX[closestPoint]);
    print(" y = " + PointsY[closestPoint]);
    println();
  } else {
    SelectedPoint1 = SelectedPoint2 = -1;    
  }
}

public void keyTyped() {
  if (key == 'n') {
    // create new point in between the two selected points    
    int minSelectedPoint = min(SelectedPoint1, SelectedPoint2);
    int maxSelectedPoint = max(SelectedPoint1, SelectedPoint2);
    if ((minSelectedPoint % 3 == 0) && (minSelectedPoint + 3 == maxSelectedPoint)) {     
      float newPointX = bezierPoint(PointsX[minSelectedPoint], PointsX[minSelectedPoint + 1],
                                    PointsX[minSelectedPoint + 2], PointsX[minSelectedPoint + 3], 0.5f);
      float newPointY = bezierPoint(PointsY[minSelectedPoint], PointsY[minSelectedPoint + 1],
                                    PointsY[minSelectedPoint + 2], PointsY[minSelectedPoint + 3], 0.5f);
      // generate two handle points
      float dX = 0.2f * (PointsX[maxSelectedPoint] - PointsX[minSelectedPoint]);
      float dY = 0.2f * (PointsY[maxSelectedPoint] - PointsY[minSelectedPoint]);      
      float handle1X = newPointX - dX;
      float handle1Y = newPointY - dY;
      float handle2X = newPointX + dX;
      float handle2Y = newPointY + dY;
      float[] newPointsX = new float[]{ handle1X, newPointX, handle2X };
      float[] newPointsY = new float[]{ handle1Y, newPointY, handle2Y };
      PointsX = splice(PointsX, newPointsX, minSelectedPoint + 2);
      PointsY = splice(PointsY, newPointsY, minSelectedPoint + 2);
    }
  } else if (key == 'd') {
    // delete selected point
    if (SelectedPoint1 % 3 == 0 && // curve point
        SelectedPoint1 != 0 && SelectedPoint1 != PointsX.length - 1 && // not the end points
        SelectedPoint2 == -1) {
      PointsX = concat(subset(PointsX, 0, SelectedPoint1 - 1),
                       subset(PointsX, SelectedPoint1 + 2, PointsX.length - SelectedPoint1 - 2));
      PointsY = concat(subset(PointsY, 0, SelectedPoint1 - 1),
                       subset(PointsY, SelectedPoint1 + 2, PointsY.length - SelectedPoint1 - 2));
    }
  }
}

public int GetClosestPointToMouse() {
  float closestDist = 1e4f;
  int closestPoint = -1;
  for (int i = 0; i < PointsX.length; ++i) {
    float currDist = dist(mouseX, mouseY, PointsX[i], PointsY[i]);
    if (currDist < PointDiameter / 2 && currDist < closestDist) {
      closestDist = currDist;
      closestPoint = i;
    }
  }
  return closestPoint;
}

public void mouseDragged() {
  if (Dragging) {
    float dx = mouseX - DraggedPointStartX;
    float dy = mouseY - DraggedPointStartY;
    PointsX[DraggedPoint] += dx;
    PointsY[DraggedPoint] += dy;
    DraggedPointStartX = mouseX;
    DraggedPointStartY = mouseY;
    if (keyPressed && key == CODED && keyCode == CONTROL) { // move everything
      PointsX[DraggedPoint] -= dx;
      PointsY[DraggedPoint] -= dy;
      for (int i = 0; i < PointsX.length; ++i) {
        PointsX[i] += dx;
        PointsY[i] += dy;
      }
    } else if (DraggedPoint % 3 != 0) { // if dragging a handle point           
      // find the other handle point
      int otherPoint = DraggedPoint % 3 == 1 ? DraggedPoint - 2
                                             : DraggedPoint + 2;
      // find the curve point
      int cPoint = DraggedPoint % 3 == 1 ? DraggedPoint - 1
                                         : DraggedPoint + 1;
      if (otherPoint >= 0 && otherPoint < PointsX.length) {
        PointsX[otherPoint] = PointsX[cPoint] * 2 - PointsX[DraggedPoint];
        PointsY[otherPoint] = PointsY[cPoint] * 2 - PointsY[DraggedPoint];
      }
    } else { // if dragging a curve point
      // find the two handle points
      int handlePoint1 = DraggedPoint - 1;
      int handlePoint2 = DraggedPoint + 1;
      if (handlePoint1 >= 0 && handlePoint1 < PointsX.length) {
        PointsX[handlePoint1] += dx;
        PointsY[handlePoint1] += dy;
      }
      if (handlePoint2 >= 0 && handlePoint2 < PointsX.length) {
        PointsX[handlePoint2] += dx;
        PointsY[handlePoint2] += dy;
      }
    }
  }
}

public void mouseReleased() {
  if (Dragging) {
    Dragging = false;
  }
}

public void draw() {
  background(255);
  drawAxes();
  drawPoints();
  drawLines();
  drawCurve();
}

public void drawAxes() {
  pushMatrix();
    translate(CenterX, CenterY);
    strokeWeight(2);
    stroke(255, 0, 0);
    line(-WindowWidth, 0, WindowWidth, 0);
    stroke(0, 0, 255);
    line(0, -WindowHeight, 0, WindowHeight);
  popMatrix();
}

public void drawPoints() {
  for (int i = 0; i < PointsX.length; ++i) {
    if (i % 3 == 0) {
      stroke(0, 255, 0);
      if (SelectedPoint1 == i || SelectedPoint2 == i) {
        stroke(0, 0, 255);
      }
    } else {
      stroke(255, 0, 255);
      if (SelectedPoint1 == i || SelectedPoint2 == i) {
        stroke(0, 0, 255);
      }
    }
    ellipse(PointsX[i], PointsY[i], PointDiameter, PointDiameter);    
  }
}

public void drawLines() {
  strokeWeight(1);
  stroke(127);
  for (int i = 1; i < PointsX.length; i += 3) {
    line(PointsX[i], PointsY[i], PointsX[i - 1], PointsY[i - 1]);
    line(PointsX[i + 1], PointsY[i + 1], PointsX[i + 2], PointsY[i + 2]);
  }
}

public void drawCurve() {
  noFill();  
  strokeWeight(2);
  stroke(0);
  beginShape();
    vertex(PointsX[0], PointsY[0]);
    for (int i = 1; i < PointsX.length; i += 3) {
      bezierVertex(PointsX[i], PointsY[i], PointsX[i + 1], PointsY[i + 1], PointsX[i + 2], PointsY[i + 2]);
    }
  endShape();
}
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "Week6_Tool" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
